# $Id: importNOS_CurRefs.rb 7446 2021-01-07 13:48:29Z flaterco $
# Functions for importNOS.rb:
# Current reference station import

def importCurRefs(db)

  stateMap = getOldStateMap(db)

  # The undocumented MDAPI (1.0) geogroup children query (getting stations)
  # does not work for currents.  In lieu of that, geogroups are parsed out of
  # the old style station list web pages by parse-oldcurlists.rb and left in
  # the table currents_geogroups.  This needn't be done if there are no new
  # stations.  getCurrentsGeogroups returns an empty hash if the table is not
  # present.
  currentsGeogroups = getCurrentsGeogroups(db)

  # Report missing/extra stations.
  oldcurlistsids = Set.new(currentsGeogroups.keys)
  currentsjsonsids = Set.new

  # Pass 1.
  # There will be no harcon jsons for any sids not in currents.json, but the
  # bins are a superset of those listed in currents.json.  So first, compile
  # metadata by sid.  Note that we won't have mean directions for the bins
  # that are missing from currents.json.  (We do get azimuth.)
  currents = readCurJson()
  currentsmd = Hash.new
  binsmd = Hash.new
  currents.each {|csta|

    # Avoid adding code that isn't exercised.
    raise "Notices are not null anymore" unless csta["notices"].nil?
    raise "Disclaimers are not null anymore" unless csta["disclaimers"].nil?

    sid = csta["id"]
    currentsjsonsids.add(sid)

    # type is H, S, W, or Y.
    #   H = harmonic
    #   S = subordinate
    #   W = weak and variable with null offsets
    #   Y = undocumented ?  Is that like X only 1 more?
    # 2020-12 type Y has gone away.
    type = csta["type"]
    raise "Unknown type #{type}" unless "HSW".include?(type)
    next if type != "H"

    # Not expecting hits on ref stations, but may as well check.
    if isMaybeNotFreeCurrent(sid)
      print "** Skipped ", sid, " (non-US)\n"
      next
    end

    if currentsmd.include?(sid)
      # All unchecked assumptions shall be violated.
      if currentsmd[sid][:meridian] != csta["timezone_offset"].to_i
        raise "Currentsmd meridian mismatch"
      elsif currentsmd[sid][:lat] != csta["lat"]
        raise "Currentsmd lat mismatch"
      elsif currentsmd[sid][:lng] != csta["lng"]
        raise "Currentsmd lng mismatch"
      elsif currentsmd[sid][:cname] != csta["name"]
        raise "Currentsmd cname mismatch"
      end
    else
      cname = csta["name"]

      # As of 2020-12, the only western Aleutians in currents.json are Rat
      # Islands, and they correctly show lng > 0.
      lat = csta["lat"]
      lng = csta["lng"]

      # "timezone_offset" is hours E of GMT.
      # -4, -5, -6, -8, -9, or -10 (as a char string, for some reason)
      # tides.json instead has "timemeridian" and "timezonecorr".
      claimed_meridian = csta["timezone_offset"].to_i

      begin
        statename, region, subregion = guessCurrentsGeogroups(currentsGeogroups, sid)
      rescue
        print "** #{lat} #{lng} #{cname}\n"
        raise
      end
      raise "Geogroup contained a null state" if statename.nil?
      raise "Need StateCode for #{statename}" unless StateCode.include?(statename)
      state = StateCode[statename]

      # currents.json has no state field and geogroups states are unreliable.
      # Hunt for errors.
      dumbstate, dist = guessStateDumb(stateMap, lat, lng)
      raise "State guess is nil" if dumbstate.nil?
      if state != dumbstate
=begin
        print "** Possible wrong state, #{sid} #{lat} #{lng} #{cname}\n"
        print "   Geogroup state #{state}\n"
        print "   Guessed #{dumbstate} (d=#{"%.0e" % dist})\n"
=end
        # FIXME, one-time code.
        # 2020-12 manually arbitrated 55 flagged suspicious cases.
        # Geogroup - 12
        # Guess - 43
        unless [
          "ACT7351", "CHB0303", "LIS1003", "LIS1032",
          "MOB1104", "NYH1906", "NYH1908", "NYH1928",
          "PIR0702", "PIR0705", "PIR0706", "kb0201"].include?(sid)
          state = dumbstate
          raise "No StateName for #{state}" unless StateName.include?(state)
          statename = StateName[state]
        end
      end

      # As of 2020-12, currents from MDAPI are all assigned to 21 states and
      # Puerto Rico, notwithstanding a few WA-BC border cases.  guessTimezone
      # needs the name parameter only for FM, MH, and Pacific atolls.
      tz = guessTimezone(state, cname, lat, lng)
      guessed_meridian = TZmerids[tz]
      if guessed_meridian != claimed_meridian
        print "** Inconsistent meridian: #{sid} #{cname}\n"
        print "  timezone_offset in JSON = #{claimed_meridian}\n"
        print "  Meridian of guessed time zone = #{guessed_meridian}\n"
      end

      # Stash the hash.
      currentsmd[sid] = {
        cname: cname,
        lat: lat,
        lng: lng,
        meridian: claimed_meridian,
        geogroups: [statename, region, subregion],
        state: state,
        tz: tz
      }
    end

    # Store the bins.
    sidp = sid + "_" + csta["currbin"].to_s
    # The offsets record is null and bogus except that it's where the mean
    # directions are reported.  Azimuth in the harcon record is close to
    # maxdir but not the same.
    offsets = csta["currentpredictionoffsets"]
    binsmd[sidp] = {
      depth: csta["depth"],             # can be null
      depthType: csta["depthType"],
      maxdir: offsets["meanFloodDir"],  # degrees true
      mindir: offsets["meanEbbDir"]     # degrees true
    }
  }

  setdif2 = oldcurlistsids-currentsjsonsids
  unless setdif2.empty?
    print "** Sids in oldcurlists that are not in currents.json:  #{setdif2.to_a}\n"
  end

  # 2020-12 patch for CFR1609.
  if binsmd["CFR1609_3"][:depth]  == 14 and
     binsmd["CFR1609_15"][:depth] == 14 and
     binsmd["CFR1609_25"][:depth] == 14
    binsmd["CFR1609_3"][:depth] =
      binsmd["CFR1609_15"][:depth] =
      binsmd["CFR1609_25"][:depth] = Float::NAN
  end

  # Pass 2, add the stations.
  # Enumerate sids from currentsmd, bins from the harcon.
  aliases = getConstituentAliases(db)
  currentsmd.each_pair {|sid, md_sta|
    bins = getCurrentConstForSid(db, aliases, sid)
    if bins.nil?
      print "** #{sid} missing harcon file\n"
      raise "missing harcon file"
    elsif bins.empty?
      print "** #{sid} harcon is null; skipping it.\n"
    else
      bins.each_pair {|bin_num, harcon_bin|
        sidp = sid + "_" + bin_num.to_s
        md_bin = binsmd[sidp]
	comments = nil

        # Depths have problems.  md_bin[:depth] matches the name on the web
        # site but is often missing.  harcon_bin[:depth] often seems to be of
        # a different depth type but there's no indication.  Sometimes
        # harcon_bin[:depth] is too large to be depth in feet or meters and
        # its relationship to md_bin[:depth] is mysterious.  Even when they
        # agree, they are rounded differently.

	# Nothing I can do but document the problem in comments.  What could
	# I say in a note that won't just confuse?

        if md_bin.nil? or md_bin[:depth].nil?
          depth = harcon_bin[:depth]
          comments = "No depth was provided in the stations list.  Depth is as specified in the\nharcon record.  Unknown whether it is below surface or chart datum."
	elsif md_bin[:depth].nan?
          depth = harcon_bin[:depth]
          comments = "Depth provided in the stations list was duplicated for all bins.  Depth is as\nspecified in the harcon record.  Unknown whether it is below surface or chart\ndatum."
        else
          depth = md_bin[:depth]
          comments = "Depth is as provided in the stations list.\n"
          case md_bin[:depthType]
          when "B"
            comments += "Depth is below chart datum."
          when "S"
            comments += "Depth is below surface."
          when "U"
            comments += "Unknown whether it is below surface or chart datum."
          else
            raise "Unknown depth type"
          end
          if (md_bin[:depth]-harcon_bin[:depth]).abs > 1
            comments = append(comments, "** Depth specified in the harcon record is #{harcon_bin[:depth]} ft.")
          end
        end

        name = generateName(md_sta[:cname], depth, md_sta[:geogroups], "Current")
        raise "Name in use" if nameInUse(db, name)

        mindir = maxdir = nil
        unless md_bin.nil?
          mindir = md_bin[:mindir]
          maxdir = md_bin[:maxdir]
        end
        if maxdir.nil?
          maxdir = harcon_bin[:azimuth]
          comments = append(comments, "No mean flood direction provided.  Used azimuth from harcon record.")
        end

	# Add data set record.  datumkind should be left null.
	units = "knots"    # Future support SI units
	legal = nil        # Saved for future use
	notes = nil        # Saved for future use
	xfields = "Credit:NOAA data processed by David Flater for XTide\n https://flaterco.com/xtide/"
	meridian = "#{md_sta[:meridian]}:00"
	res = db.exec("insert into data_sets (name, original_name, station_id_context, station_id, lat, lng, timezone, state, country, units, source, legalese, notes, comments, restriction, xfields, min_dir, max_dir, meridian, datum) values ($1, $2, 'NOS', $3, $4, $5, $6, $7, 'USA', $8, 'CO-OPS Metadata API', $9, $10, $11, 'Public domain', $12, $13, $14, $15, $16) returning index", [name, md_sta[:cname], sidp, md_sta[:lat], md_sta[:lng], md_sta[:tz], md_sta[:state], units, legal, notes, comments, xfields, mindir, maxdir, meridian, harcon_bin[:datum]])
	index = res.getvalue(0,0).to_i

	# Add constants.
        const_added = 0
	harcon_bin[:constants].each_pair {|constname, const|
          # De-aliasing of constname was already done.
          amp, phase = const[:amp], const[:phase]
	  next if amp < 0.00005
	  phase = 0 if phase == 360  # This happens
	  raise "Phase out of range" if phase < 0 or phase >= 360 # Also happens...
	  res = db.exec("insert into constants values ($1, $2, $3, $4)",
			 [index, constname, phase, amp])
	  const_added += 1
	}
        if const_added.zero?
	  print "** No harmonic constants added for #{sidp} #{name}\n"
          raise "Broken station"
        end
      }
    end
  }
end
