# $Id: importNOS_CurSubs.rb 7446 2021-01-07 13:48:29Z flaterco $
# Functions for importNOS.rb:
# Current subordinate station import

def importCurSubs(db)

  # FIXME
  # 2020-12, including the new current refs and few new tide refs (which have
  # already been validated) in the mix in a desperate bid to reduce the
  # number of stations needing manual arbitration of state.  (This affected
  # only 13 stations out of 208.)
  stateMap = getMergedStateMap(db)

  # The undocumented MDAPI (1.0) geogroup children query (getting stations)
  # does not work for currents.  In lieu of that, geogroups are parsed out of
  # the old style station list web pages by parse-oldcurlists.rb and left in
  # the table currents_geogroups.  This needn't be done if there are no new
  # stations.  getCurrentsGeogroups returns an empty hash if the table is not
  # present.
  currentsGeogroups = getCurrentsGeogroups(db)

  # For state assignments that were already arbitrated, hash from sid to
  # state code.
  arbitratedStates = Hash.new
  # Preload one exception where the correct answer wasn't even on the menu.
  arbitratedStates["ACT4366"] = "DE"

  currents = readCurJson()
  currents.each {|csta|

    raise "Notices are not null anymore" unless csta["notices"].nil?
    raise "Disclaimers are not null anymore" unless csta["disclaimers"].nil?

    # type is H, S, W, or Y.
    #   H = harmonic
    #   S = subordinate
    #   W = weak and variable with null offsets
    #   Y = undocumented ?  Is that like X only 1 more?
    # 2020-12 type Y has gone away.
    type = csta["type"]
    raise "Unknown type #{type}" unless "HSW".include?(type)
    next if type != "S"

    sid = csta["id"]
    sidp = sid + "_" + csta["currbin"].to_s
    cname = csta["name"]
    comments = notes = legal = nil

    if isMaybeNotFreeCurrent(sid)
      print "Skipped ", sidp, " ", cname, " (non-US)\n"
      next
    end

    offsets = csta["currentpredictionoffsets"]
    refsid = offsets["refStationId"]
    raise "Current sub with no refsid" if refsid.nil? or refsid.empty?
    refsidp = refsid + "_" + offsets["refStationBin"].to_s

    # Demand exactly one ref station with refsidp.
    # Count 0  - complain.
    # Count 1  - good.
    # Count >1 - throw.
    if !refStationExists(db, refsidp)
      temp = refsidp.upcase
      if !refStationExists(db, temp)
        print "** Reference station #{refsidp} for sub #{sidp} does not exist\n"
	next
      else
        refsidp = temp
      end
    end
    refIndex = getRefIndex(db,refsidp)  # Throws on count != 1

    # As of 2020-12, the only western Aleutians in currents.json are Rat
    # Islands, and they correctly show lng > 0.
    lat = csta["lat"]
    lng = csta["lng"]

    mindir = offsets["meanEbbDir"]    # degrees true
    maxdir = offsets["meanFloodDir"]  # degrees true
    mint   = offsets["mecTimeAdjMin"] # minutes
    maxt   = offsets["mfcTimeAdjMin"] # minutes
    sbft   = offsets["sbfTimeAdjMin"] # minutes
    sbet   = offsets["sbeTimeAdjMin"] # minutes
    minlm  = offsets["mecAmpAdj"]     # multiplier
    maxlm  = offsets["mfcAmpAdj"]     # multiplier
    raise "Nil offsets" if mint.nil? or maxt.nil? or sbft.nil? or
      sbet.nil? or minlm.nil? or maxlm.nil?

    fixednulls = false
    minlm, fixednulls = [0.001, true] if minlm <= 0
    maxlm, fixednulls = [0.001, true] if maxlm <= 0
    if fixednulls
      print "Fixed null multipliers at ", sidp, " ", cname, "\n"
      notes = append(notes, "Velocities for this station should be disregarded--use times only.")
      legal = "IGNORE VELOCITIES"
    end

    begin
      statename, region, subregion = guessCurrentsGeogroups(currentsGeogroups, sid)
    rescue
      print "** #{lat} #{lng} #{cname}\n"
      raise
    end
    raise "Geogroup contained a null state" if statename.nil?
    raise "Need StateCode for #{statename}" unless StateCode.include?(statename)
    state = StateCode[statename]    # Nearly worthless...

    # currents.json has no state field and geogroups states are unreliable.
    # Hunt for errors.
    # FIXME, one-time code.
    if arbitratedStates.include?(sid)
      state = arbitratedStates[sid]
      raise "No StateName for #{state}" unless StateName.include?(state)
      statename = StateName[state]
    else
      # In 11 identified cases, keep the geogroup state.
      unless [
        "ACT2561", "ACT2591", "ACT3656", "ACT4091", "ACT4236", "ACT4371",
        "ACT4416", "ACT4441", "ACT0726", "ACT4471", "ACT5751"].include?(sid)
	# In all other cases, dumb guess >= geogroup state.  (Often they are
	# the same.)
	dumbstate, dist = guessStateDumb(stateMap, lat, lng)
	raise "State guess is nil" if dumbstate.nil?
        state = dumbstate
        raise "No StateName for #{state}" unless StateName.include?(state)
        statename = StateName[state]
      end
      arbitratedStates[sid] = state
    end

    # FIXME, remnants of the cumbersome arbitration process used in 2020.
    # Reuse as applicable depending on whether new currents appear in
    # batches.
=begin
    # If the coordinates have an exact match, don't spam me, just trust
    # the matching data.
    if dist > 0
      res = db.exec("insert into arb values ($1, $2, $3, $4, $5, $6, $7)", [sid, cname, lat, lng, state, dumbstate, nil])
      # STDERR.print "** Possible wrong state, #{sid} #{lat} #{lng} #{cname}\n"
      # STDERR.print "   Geogroup state #{state}\n"
      # STDERR.print "   Guessed #{dumbstate} (d=#{"%.0e" % dist})\n"
    end
=end

    # As of 2020-12, currents from MDAPI are all assigned to 21 states and
    # Puerto Rico, notwithstanding a few WA-BC border cases.  guessTimezone
    # needs the name parameter only for FM, MH, and Pacific atolls.
    tz = guessTimezone(state, cname, lat, lng)

    # Time warps and other meridian problems
    # (begin near-duplication of code from TideSubs)
    submerid = TZmerids[tz]
    if submerid.nil?
      print "Missing from TZmerids: ", tz, "\n"
      raise "Missing tz"
    end
    # "timezone_offset" is *usually* an empty string for sub stations.
    # Hours E of GMT.
    # -4, -5, -6, -8, -9, or -10 (as a char string, for some reason)
    # tides.json instead has "timemeridian" and "timezonecorr".
    unless csta["timezone_offset"].empty?
      claimed_meridian = csta["timezone_offset"].to_i
      if submerid != claimed_meridian
        print "** Inconsistent meridian: #{sid} #{cname}\n"
        print "  timezone_offset in JSON = #{claimed_meridian}\n"
        print "  Meridian of guessed time zone = #{submerid}\n"
      end
    end
    reftz = getRefTimezone(db, refsidp)
    if reftz != tz
      refmerid = TZmerids[reftz]
      if refmerid.nil?
	print "Missing from TZmerids: ", reftz, "\n"
	raise "Missing tz"
      end
      if refmerid != submerid
	print "Time warp at ", sidp, " ", cname, "\n"
	print "  Ref ", refsidp, " is ", reftz, " (merid ", refmerid, ")\n"
	print "  Sub is ", tz, " (merid ", submerid, ")\n"
	adjustMin = (refmerid - submerid) * 60
	mint += adjustMin
	maxt += adjustMin
	sbft += adjustMin
	sbet += adjustMin
	notes = append(notes, "Reference is in time zone " + reftz +
		       "; adjusted offsets " + (refmerid-submerid).to_s +
		       " hr\nto undo LST compensation.")
      end
    end

    depth = csta["depth"]
    if depth.nil?
      comments = append(comments, "No depth was provided in the stations list.")
    else
      comments = append(comments, "Depth is as provided in the stations list.")
      case csta["depthType"]
      when "B"
	comments = append(comments, "Depth is below chart datum.")
      when "S"
	comments = append(comments, "Depth is below surface.")
      when "U"
	comments = append(comments, "Unknown whether it is below surface or chart datum.")
      else
	raise "Unknown depth type"
      end
    end

    name = generateName(cname, depth, [statename, region, subregion], "Current")
    raise "Name in use" if nameInUse(db, name)

    units = "knots"    # Future support SI units
    country = "USA"    # Saved for future use
    xfields = "Credit:NOAA data processed by David Flater for XTide\n https://flaterco.com/xtide/"
    mint = mint.to_s + " minutes"
    maxt = maxt.to_s + " minutes"
    sbft = sbft.to_s + " minutes"
    sbet = sbet.to_s + " minutes"
    res = db.exec("insert into data_sets (name, original_name, station_id_context, station_id, lat, lng, timezone, state, country, units, source, legalese, notes, comments, restriction, xfields, ref_index, min_time_add, min_level_multiply, max_time_add, max_level_multiply, flood_begins, ebb_begins, min_dir, max_dir) values ($1, $2, 'NOS', $3, $4, $5, $6, $7, $8, $9, 'CO-OPS Metadata API', $10, $11, $12, 'Public domain', $13, $14, $15, $16, $17, $18, $19, $20, $21, $22)", [name, cname, sidp, lat, lng, tz, state, country, units, legal, notes, comments, xfields, refIndex, mint, minlm, maxt, maxlm, sbft, sbet, mindir, maxdir])
  }
end
